
/******************************************************************************
* MODULE     : edit_session.gen.cc
* DESCRIPTION: input and output handling for programming sessions
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "Process/edit_process.hh"
#include "list.hh"
#include "connect.hh"
#include "analyze.hh"

extern bool eval_scheme_func_tree_tree (string f, tree t, tree& r);
extern string tree_to_verbatim (tree t);

/******************************************************************************
* Other useful subroutines
******************************************************************************/

static bool
is_empty (tree t) {
  if (is_atomic (t)) return (t == "");
  if (is_expand (t, "math", 1)) return is_empty (t[1]);
  if (is_document (t) || is_concat (t)) {
    int i;
    for (i=0; i<N(t); i++)
      if (!is_empty (t[i])) return false;
    return true;
  }
  return false;
}

/******************************************************************************
* Routines for sessions
******************************************************************************/

void
edit_process_rep::make_session (string lan, string session) {
  if ((lan != "scheme") && (lan != "shell")) init_extra_style (lan, true);
  typeset_preamble ();

  /* These two lines will become obsolete */
  if (make_return_after ()) return;
  if (make_return_before ()) return;
  /* -------------------------------------*/

  tree body (DOCUMENT, "");
  path p (3, path (0, 0));
  insert_tree (tree (EXPAND, "session", lan, session, body), p);
  if (lan == "shell") start_input ();
  else if (connection_declared (lan)) connect ();
  else start_input ();
}

void
edit_process_rep::start_input () {
  path p;
  bool needs_return= false;
  if (!nil (p= search_upwards_expand ("input"))) {
    go_to (p * 1);
    needs_return= true;
  }
  else if (!nil (p= search_upwards_expand ("output"))) {
    tree st= subtree (et, p);
    if ((N(st) == 2) && is_empty (st [1])) {
      cut (p);
      remove_backwards ();
    }
    else go_to (p * 1);
    needs_return= true;
  }

  string lan   = get_env_string (PROG_LANGUAGE);
  string session= get_env_string (THIS_SESSION);
  tree   prompt = "";
  tree   input  = "";

  if (connection_declared (lan)) {
    prompt = connection_read (lan, session, 0, "prompt");
    input  = connection_read (lan, session, 0, "input");
  }
  while (is_document (prompt))
    if (N(prompt) == 0) prompt= "";
    else prompt= prompt[0];
  if (prompt == "") prompt= lan * "] ";
  if (!is_document (input)) input= tree (DOCUMENT, input);
  if (math_input) input= tree (VAR_EXPAND, "math", input);

  path q = path_up (tp, 2);
  int  i = last_item (path_up (tp));
  tree st= subtree (et, q);
  if (is_document (st) &&
      (i+1 < N(st)) &&
      is_expand (st[i+1], "input", 2))
    {
      if (is_empty (input)) {
	tree old= copy (st[i+1][2]);
	if (!(math_input ^ is_expand (old, "math", 1))) input= old;
      }
      assign (q * (i+1), tree (EXPAND, "input", prompt, input));
      go_to (q * path (i+1, 2, end (input)));
    }
  else {
    if (needs_return) insert_return ();
    insert_tree (tree (EXPAND, "input", prompt, input), path (2, end (input)));
  }
  if (message_l != "") {
    set_message (message_l, message_r);
    message_l= "";
    message_r= "";
  }
  else set_message ("", "");
}

void
edit_process_rep::start_output () {
  path p;
  bool needs_return= false;
  if (!nil (p= search_upwards_expand ("input"))) {
    go_to (p * 1);
    needs_return= true;
  }
  else if (!nil (p= search_upwards_expand ("output"))) {
    go_to (p * 1);
    needs_return= true;
  }

  path q = path_up (tp, 2);
  int  i = last_item (path_up (tp));
  tree st= subtree (et, q);
  if (is_document (st) &&
      (i+1 < N(st)) &&
      is_expand (st[i+1], "output", 1))
    {
      assign (q * (i+1), tree (EXPAND, "output", tree (DOCUMENT, "")));
      go_to (q * path (i+1, 1, 0, 0));
    }
  else {
    if (needs_return) insert_return ();
    insert_tree (tree (EXPAND, "output", tree (DOCUMENT, "")),
		 path (1, path (0, 0)));
  }
}

void
edit_process_rep::session_message (string l, string r) {
  message_l= l;
  message_r= r;
}

void
edit_process_rep::session_use_math_input (bool flag) {
  if (math_input != flag) {
    math_input= flag;
    path p= search_upwards_expand ("input");
    if (nil (p)) return;
    tree input (DOCUMENT, "");
    path q (0, 0);
    if (math_input) {
      input= tree (VAR_EXPAND, "math", input);
      q= path (1, q);
    }
    assign (p * 2, input);
    go_to (p * path (2, q));
  }
}

bool
edit_process_rep::session_is_using_math_input () {
  return math_input;
}

/******************************************************************************
* Cursor movement inside sessions
******************************************************************************/

void
edit_process_rep::session_var_go_up () {
  path p= search_upwards_expand ("session");
  if (nil (p) || (N(p)+2 > N(tp))) return;
  tree st= subtree (et, p);
  if ((N(st) != 4) || (tp[N(p)]!=3) || (!is_document (st[3]))) return;
  int i= tp[N(p)+1];
  p= p * 3;
  st= st[3];
  for (i--; i>=0; i--)
    if (is_expand (st[i], "input", 2)) {
      go_to (p * path (i, 2, end (st[i][2])));
      select_from_cursor_if_active ();
      return;
    }
}

void
edit_process_rep::session_var_go_down () {
  path p= search_upwards_expand ("session");
  if (nil (p) || (N(p)+2 > N(tp))) return;
  tree st= subtree (et, p);
  if ((N(st) != 4) || (tp[N(p)]!=3) || (!is_document (st[3]))) return;
  int i= tp[N(p)+1];
  p= p * 3;
  st= st[3];
  for (i++; i<N(st); i++)
    if (is_expand (st[i], "input", 2)) {
      go_to (p * path (i, 2, end (st[i][2])));
      select_from_cursor_if_active ();
      return;
    }
}

void
edit_process_rep::session_go_left () {
  path p= search_upwards_expand ("input");
  if (nil (p)) return;
  int i= tp[N(p)];
  path old_tp= tp;
  go_left ();
  p= search_upwards_expand ("input");
  if (nil (p) || ((tp[N(p)] != 2) && (tp[N(p)] != i))) go_to (old_tp);
  select_from_cursor_if_active ();
}

void
edit_process_rep::session_go_right () {
  path p= search_upwards_expand ("input");
  if (nil (p)) return;
  int i= tp[N(p)];
  path old_tp= tp;
  go_right ();
  p= search_upwards_expand ("input");
  if (nil (p) || ((tp[N(p)] != 2) && (tp[N(p)] != i))) go_to (old_tp);
  select_from_cursor_if_active ();
}

void
edit_process_rep::session_go_up () {
  path p= search_upwards_expand ("input");
  if (nil (p)) return;
  int i= tp[N(p)];
  path old_tp= tp;
  go_up ();
  p= search_upwards_expand ("input");
  if (nil (p) || ((tp[N(p)] != 2) && (tp[N(p)] != i))) {
    go_to (old_tp);
    session_var_go_up ();
  }
  select_from_cursor_if_active ();
}

void
edit_process_rep::session_go_down () {
  path p= search_upwards_expand ("input");
  if (nil (p)) return;
  int i= tp[N(p)];
  path old_tp= tp;
  go_down ();
  p= search_upwards_expand ("input");
  if (nil (p) || ((tp[N(p)] != 2) && (tp[N(p)] != i))) {
    go_to (old_tp);
    session_var_go_down ();
  }
  select_from_cursor_if_active ();
}

void
edit_process_rep::session_go_page_up () {
  int i;
  for (i=0; i<5; i++)
    session_var_go_up ();
}

void
edit_process_rep::session_go_page_down () {
  int i;
  for (i=0; i<5; i++)
    session_var_go_down ();
}

void
edit_process_rep::session_remove_backwards () {
  path p= search_upwards_expand ("math");
  if (nil (p)) {
    p= search_upwards_expand ("input");
    if (nil (p) || (tp == start (et, p * 2))) return;
  }
  else if (tp == start (et, p * 1)) return;
  remove_backwards ();
}

void
edit_process_rep::session_remove_forwards () {
  path p= search_upwards_expand ("math");
  if (nil (p)) {
    p= search_upwards_expand ("input");
    if (nil (p) || (tp == end (et, p * 2))) return;
  }
  else if (tp == end (et, p * 1)) return;
  remove_forwards ();
}

/******************************************************************************
* Manipulating the input line
******************************************************************************/

string
edit_process_rep::session_get_input () {
  return "";
}

int
edit_process_rep::session_get_input_pos () {
  return 0;
}

void
edit_process_rep::session_set_input (string s) {
  (void) s;
}

void
edit_process_rep::session_set_input_pos (int pos) {
  (void) pos;
}

/******************************************************************************
* Tab completion
******************************************************************************/

static string cursor_symbol ("[tmcursor]");

static tree
put_cursor (tree t, path p) {
  if (is_atomic (t)) {
    string s= t->label;
    return s (0, p->item) * cursor_symbol * s (p->item, N(s));
  }
  else {
    if (p == path (0)) return tree (CONCAT, cursor_symbol, t);
    else if (p == path (1)) return tree (CONCAT, t, cursor_symbol);
    else {
      int i, n= N(t);
      tree u (t, n);
      for (i=0; i<n; i++)
	if (i == p->item) u[i]= put_cursor (t[i], p->next);
	else u[i]= t[i];
      return u;
    }
  }
}

bool
edit_process_rep::session_complete_try () {
  path p= search_upwards_expand ("input");
  if (nil (p)) return false;
  tree st= subtree (et, p);
  if ((N(tp) <= N(p)) || (tp[N(p)] != 2)) return false;
  tree t= put_cursor (st[2], tail (tp, N(p)+1));
  // cout << t << LF;

  string lan= get_env_string (PROG_LANGUAGE);
  string ses= get_env_string (THIS_SESSION);
  if (is_func (t, DOCUMENT, 1)) t= t[0];
  if (is_expand (t, "math", 1)) {
    tree in= tuple (lan, t[1]), out;
    if (!eval_scheme_func_tree_tree ("cas-math-input", in, out)) t= out;
  }
  // cout << t << LF;

  string s= tree_to_verbatim (t);
  int pos= search_forwards (cursor_symbol, s);
  if (pos == -1) return false;
  s= s (0, pos) * s (pos + N(cursor_symbol), N(s));
  // cout << s << ", " << pos << LF;

  string body=
    "(complete \"" * escape_quotes (s) *
    "\" " * as_string (pos) *")";
  string send=
    string (DATA_BEGIN) * "special:" *
    escape_tmprot (body) * string (DATA_END);
  // cout << send << LF;

  tree r= connection_eval (lan, ses, send);
  if (is_func (r, DOCUMENT, 1)) r= r[0];
  // cout << r << LF;

  if (!is_tuple (r)) return false;
  int i, n= N(r);
  string prefix;
  array<string> compls;
  for (i=0; i<n; i++)
    if (is_atomic (r[i])) {
      string l= r[i]->label;
      if (is_quoted (l)) l= unquote (l);
      if (prefix == "") prefix= l;
      else compls << l;
    }
  // cout << prefix << ", " << compls << LF;

  if ((prefix == "") || (N(compls) == 0)) return false;
  complete_start (prefix, compls);
  return true;
}
